/* Copyright © 2010 Brandon L Black <blblack@gmail.com> and Jay Reitz <jreitz@gmail.com>
 *
 * This file is part of gdnsd.
 *
 * gdnsd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * gdnsd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gdnsd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _GDNSD_DNSPACKET_H
#define _GDNSD_DNSPACKET_H

#include "config.h"
#include "gdnsd.h"
#include "ltree.h"
#if DNSCURVE
#  include "dnscurve_data.h"
#endif

// dnspacket-layer statistics, per-thread
typedef struct {
  bool is_udp;

  // Per-protocol stats
  union {
    struct { // UDP stats
      satom_t sendfail;
      satom_t tc;
      satom_t edns;
      satom_t edns_big;
      satom_t edns_tc;
    } udp;
    struct { // TCP stats
      satom_t recvfail;
      satom_t recvsize;
      satom_t sendfail;
    } tcp;
  } p;

  // DNS layer stats, first 5 directly correspond to RCODEs
  // All 6, summed, represent the total count
  //  of requests received by the DNS layer.  Note that
  //  some of the earlier UDP/TCP-specific failures never make it to
  //  to the DNS layer.
  satom_t noerror;
  satom_t refused;
  satom_t nxdomain;
  satom_t notimp;
  satom_t badvers;
  satom_t dropped; // no response sent at all, horribly badly formatted

  // Count of requests over IPv6.  The only valid relation to other stats
  // is that you could compare it to the 6-stat sum above for a percentage
  satom_t v6;

  // The lack of "if DNSCURVE" is intentional, I'd rather the
  // core stats output be consistent for other tools to parse, even
  // if these values are always zero in non-DNSCURVE builds
  satom_t curve_strm;
  satom_t curve_strm_fail;
  satom_t curve_txt;
  satom_t curve_txt_fail;
  satom_t curve_cache_hit;
  satom_t curve_cache_miss;
} dnspacket_stats_t;

typedef struct {
    const uint8_t* original; // Alias to the original uncompressed dname's data (not the len byte)
    const uint8_t* comp_ptr; // where compression occured on storage (could be off the end if uncompressed)
    unsigned int stored_at; // offset this name was first stored to in the packet, possibly partially compressed
} comptarget_t;

// DNS request context.  You must have a unique
//  one of these for each thread that might call
//  into process_dns_query().
typedef struct {
#if DNSCURVE
    uint8_t* curve_txt_buf;
    nonce64_state_t* n64_state;
    unsigned num_curve_keys;
    io_curvekey_t* curve_keys;
    kc_t* keycache;
#endif

    const ltree_rrset_addr_t* glue_addrs[32];
    unsigned int glue_nameptrs[32];

    const ltree_rrset_t* additional_rrsets[256];
    unsigned int additional_nameptrs[256];

    // Compression offsets, these are one per domainname in the whole
    //  packet.  Fully compressed names are not added, so this is really
    //  the number of unique domainnames in a response packet, so 255
    //  should be plenty.
    comptarget_t comptargets[256];
    unsigned int comptarget_count; // unique domainnames stored to the packet, including the original question

    // This is an offset into a query name where the zone name
    //   started.  e.g. for querying foo.example.com , if we
    //   find an authoritative name in example.com, auth_depth
    //   will be 4 (skipping \3foo).  For foo.subz.example.com,
    //   if we find the name is delegated to the subzone
    //   subz.example.com, we would still return 4 (if it were
    //   not delegated, it would be 9).
    unsigned int auth_depth;

    // The globally unique thread number for this context
    unsigned threadnum;

    // whether the thread using this context is a udp or tcp thread
    bool is_udp;

    // Max response size for this individual request, as determined
    //  by protocol type and EDNS (or lack thereof)
    unsigned int this_max_response;

    dnspacket_stats_t* stats;

    // used to pseudo-randomly rotate some RRsets (A, AAAA, NS, PTR)
    uint32_t lfsr;

    // These describe the question
    unsigned int qtype;  // Same numeric values as RFC
    unsigned int qname_comp; // compression pointer for the current query name, starts at 0xC00C, changes when following CNAME chains
    const anysin_t* client; // client info

// From this point (answer_addr_rrset) on, all of this gets reset to zero
//  at the start of each request...
    dynaddr_result_t ans_dyna; // dnspacket<->plugin comms happen here
    dyncname_result_t ans_dyncname;

    const ltree_rrset_addr_t* answer_addr_rrset;
    const ltree_rrset_srv_t* answer_srv_rrset;
    unsigned int glue_count; // refs above, not physical RRs - logical queries to resolve into 1+ RRs
    unsigned int additional_count; // refs above, not physical RRs - logical queries to resolve into 1+ RRs

    unsigned int ancount;
    unsigned int nscount;
    unsigned int arcount;
    unsigned int cname_ancount;

    // Whether this request had a valid EDNS0 optrr
    bool use_edns;

    // If this is true, the query class was CH
    bool chaos;
} dnspacket_context_t;

F_NONNULL
unsigned int process_dns_query(dnspacket_context_t* c, const anysin_t* asin, uint8_t* packet, const unsigned int packet_len);

F_MALLOC F_WUNUSED
dnspacket_context_t* dnspacket_context_new(const unsigned int this_threadnum, const bool is_udp);

void dnspacket_global_setup(void);
void dnspacket_wait_stats(void);

extern dnspacket_stats_t** dnspacket_stats;

#endif // _GDNSD_DNSPACKET_H
