/* Copyright © 011 Logitech
 *
 * This file is part of gdnsd.
 *
 * gdnsd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * gdnsd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gdnsd.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Brandon L Black <blblack@gmail.com>
 */

#define GDNSD_PLUGIN_NAME multifo

#include "config.h"
#include <gdnsd-plugin.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>

typedef struct {
    anysin_t addr;
    monio_state_t state;
} addrstate_t;

typedef struct {
    const char* name;
    addrstate_t* addrstates_v4;
    addrstate_t* addrstates_v6;
    unsigned count_v4;
    unsigned count_v6;
} mfo_resource_t;

static mfo_resource_t* resources = NULL;
static unsigned num_resources = 0;

static monio_list_t monio_list = { 0, NULL };

/*********************************/
/* Local, static functions       */
/*********************************/

static void monio_add(const char* svctype, const char* desc, const char* addr_txt, monio_state_t* state_ptr) {
    monio_list.info = realloc(monio_list.info, sizeof(monio_info_t) * (monio_list.count + 1));
    monio_info_t* m = &monio_list.info[monio_list.count++];
    m->svctype = svctype;
    m->desc = desc;
    m->addr = addr_txt;
    m->state_ptr = state_ptr;
}

F_NONNULL
static bool bad_res_opt(const char* key, unsigned klen V_UNUSED, const vscf_data_t* d V_UNUSED, void* data) {
    dmn_assert(key); dmn_assert(d); dmn_assert(data);
    log_fatal("plugin_multifo: resource '%s': bad option '%s'", (const char*)data, key);
}

F_NONNULL
static bool config_res(const char* resname, unsigned resname_len V_UNUSED, const vscf_data_t* opts, void* data) {
    dmn_assert(resname); dmn_assert(opts); dmn_assert(data);

    unsigned* residx_ptr = data;
    unsigned res = (*residx_ptr)++;
    mfo_resource_t* this_res = &resources[res];
    this_res->name = strdup(resname);

    if(vscf_get_type(opts) != VSCF_HASH_T)
        log_fatal("plugin_multifo: resource %s: value must be a hash", resname);

    const char* svctype = NULL;
    const vscf_data_t* svctype_data = vscf_hash_get_data_byconstkey(opts, "service_type", true);
    if(svctype_data) {
        if(VSCF_SIMPLE_T != vscf_get_type(svctype_data))
            log_fatal("plugin_multifo: resource %s: 'service_type' must be a string", resname);
        svctype = vscf_simple_get_data(svctype_data);
    }

    unsigned num_addrs;
    const vscf_data_t* addrs_data = vscf_hash_get_data_byconstkey(opts, "addrs", true);
    if(!addrs_data || VSCF_HASH_T != vscf_get_type(addrs_data) || !(num_addrs = vscf_hash_get_len(addrs_data)))
        log_fatal("plugin_multifo: resource %s: 'addrs' must exist and be a hash of one or more desc => IPv4 mappings", resname);

    const struct addrinfo hints = {
        .ai_flags = AI_NUMERICHOST,
        .ai_family = AF_UNSPEC,
        .ai_socktype = 0,
        .ai_protocol = 0,
        .ai_addrlen = 0,
        .ai_addr = NULL,
        .ai_canonname = NULL,
        .ai_next = NULL
    };

    // This is a cheap hack to keep the code simpler: we're allocating
    //  both addrstate arrays to the overall count, when we only really
    //  need half of that space.  Otherwise we have to store a bunch of extra
    //  state or deal with realloc issues..
    this_res->addrstates_v4 = calloc(num_addrs, sizeof(addrstate_t));
    this_res->addrstates_v6 = calloc(num_addrs, sizeof(addrstate_t));

    for(unsigned i = 0; i < num_addrs; i++) {
        const char* addr_desc = vscf_hash_get_key_byindex(addrs_data, i, NULL);
        const vscf_data_t* addr_data = vscf_hash_get_data_byindex(addrs_data, i);
        if(!vscf_is_simple(addr_data))
            log_fatal("plugin_multifo: resource %s: address %s: all addresses must be strings", resname, addr_desc);
        const char* addr_txt = vscf_simple_get_data(addr_data);

        struct addrinfo* ainfo;
        int addr_err = getaddrinfo(addr_txt, NULL, &hints, &ainfo);
        if(addr_err)
            log_fatal("plugin_multifo: resource %s: failed to parse address '%s' for '%s': %s", resname, addr_txt, addr_desc, gai_strerror(addr_err));
        if(ainfo->ai_addr->sa_family != AF_INET6 && ainfo->ai_addr->sa_family != AF_INET)
            log_fatal("plugin_multifo: resource %s: unknown address family for address '%s' for '%s': %s", resname, addr_txt, addr_desc, gai_strerror(addr_err));

        addrstate_t* as;
        if(ainfo->ai_addr->sa_family == AF_INET6) {
            as = &this_res->addrstates_v6[this_res->count_v6++];
            if(this_res->count_v6 > 64)
                log_fatal("plugin_multifo: resource %s: too many IPv6 addresses (limit 64)", resname);
        }
        else {
            as = &this_res->addrstates_v4[this_res->count_v4++];
            if(this_res->count_v4 > 64)
                log_fatal("plugin_multifo: resource %s: too many IPv4 addresses (limit 64)", resname);
        }

        memcpy(&as->addr, ainfo->ai_addr, ainfo->ai_addrlen);
        freeaddrinfo(ainfo);

        char *complete_desc = malloc(strlen(resname) + strlen(addr_desc) + 2);
        strcpy(complete_desc, resname);
        strcat(complete_desc, "/");
        strcat(complete_desc, addr_desc);
        monio_add(svctype, complete_desc, addr_txt, &as->state);
    }

    vscf_hash_iterate(opts, true, bad_res_opt, (void*)resname);
    return true;
}

/*********************************/
/* Exported callbacks start here */
/*********************************/

F_NONNULL
monio_list_t* plugin_multifo_load_config(const vscf_data_t* config) {
    dmn_assert(config);
    dmn_assert(vscf_get_type(config) == VSCF_HASH_T);

    num_resources = vscf_hash_get_len(config);
    resources = calloc(num_resources, sizeof(mfo_resource_t));
    unsigned residx = 0;
    vscf_hash_iterate(config, false, config_res, &residx);

    return &monio_list;
}

unsigned plugin_multifo_map_resource_dyna(const char* resname) {
    if(!resname)
        log_fatal("multifo plugin requires a resource name");

    for(unsigned i = 0; i < num_resources; i++)
        if(!strcmp(resname, resources[i].name))
            return i;

    log_fatal("Unknown multifo plugin resource '%s'", resname);
}

void plugin_multifo_resolve_dynaddr(unsigned threadnum V_UNUSED, unsigned resnum, const anysin_t* client V_UNUSED, dynaddr_result_t* result) {
    bool all_up = true;
    mfo_resource_t* rsrc = &resources[resnum];

    // Add any UP/DANGER IPs to the results, and track whether
    //  there were any DANGER/DOWN at all or not.
    // Note that the v4/v6 sets are treated independently.  This
    //  makes the most sense when you think through the scenarios...
 
    for(unsigned i = 0; i < rsrc->count_v4; i++) {
        const addrstate_t* as = &rsrc->addrstates_v4[i];
        const monio_state_uint_t state = monio_state_get(&as->state);
        if(state != MONIO_STATE_UP)
            all_up = false;
        if(state != MONIO_STATE_DOWN)
            result->addrs_v4[result->count_v4++] = as->addr.sin.sin_addr.s_addr;
    }

    for(unsigned i = 0; i < rsrc->count_v6; i++) {
        const addrstate_t* as = &rsrc->addrstates_v6[i];
        const monio_state_uint_t state = monio_state_get(&as->state);
        if(state != MONIO_STATE_UP)
            all_up = false;
        if(state != MONIO_STATE_DOWN)
            memcpy(&result->addrs_v6[16 * result->count_v6++], &as->addr.sin6.sin6_addr.s6_addr, 16);
    }

    // All were down, so just return all (better than empty)
 
    if(!result->count_v4) {
        for(unsigned i = 0; i < rsrc->count_v4; i++) {
            const addrstate_t* as = &rsrc->addrstates_v4[i];
            result->addrs_v4[result->count_v4++] = as->addr.sin.sin_addr.s_addr;
        }
    }

    if(!result->count_v6) {
        for(unsigned i = 0; i < rsrc->count_v6; i++) {
            const addrstate_t* as = &rsrc->addrstates_v6[i];
            memcpy(&result->addrs_v6[16 * result->count_v6++], &as->addr.sin6.sin6_addr.s6_addr, 16);
        }
    }

    // Cut TTL in half if any were in DOWN or DANGER states
    if(!all_up)
        result->ttl >>= 1;
}
