/* Copyright © 2010 Brandon L Black <blblack@gmail.com>
 *
 * This file is part of gdnsd.
 *
 * gdnsd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * gdnsd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gdnsd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define GDNSD_PLUGIN_NAME simplefo

#include "config.h"
#include <gdnsd-plugin.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

typedef struct {
    const char* name;
    uint32_t primary_ipv4;
    uint32_t secondary_ipv4;
    monio_state_t primary_state;
    monio_state_t secondary_state;
} sfo_resource_t;

static sfo_resource_t* resources = NULL;
static unsigned num_resources = 0;

static bool runtime = false;

typedef struct {
    satom_t pri;
    satom_t sec;
} sfo_stats_t;

static unsigned n_threads = 0;
static sfo_stats_t** thread_stats = NULL;

static unsigned long log_interval = 3600;

static monio_list_t monio_list = { 0, NULL };

/*********************************/
/* Local, static functions       */
/*********************************/

static void monio_add(const char* svctype, const char* resname, const char* addr_txt, monio_state_t* state_ptr) {
    monio_list.info = realloc(monio_list.info, sizeof(monio_info_t) * (monio_list.count + 1));
    monio_info_t* m = &monio_list.info[monio_list.count++];
    m->svctype = svctype;
    m->desc = resname;
    m->addr = addr_txt;
    m->state_ptr = state_ptr;
}

static bool bad_res_opt(const char* key, unsigned klen V_UNUSED, const vscf_data_t* d V_UNUSED, void* data) {
    log_fatal("plugin_simplefo: resource '%s': bad option '%s'", (const char*)data, key);
}

static bool config_res(const char* resname, unsigned resname_len V_UNUSED, const vscf_data_t* opts, void* data) {
    unsigned* residx_ptr = data;
    unsigned res = (*residx_ptr)++;
    sfo_resource_t* this_res = &resources[res];
    this_res->name = strdup(resname);

    if(vscf_get_type(opts) != VSCF_HASH_T)
        log_fatal("plugin_simplefo: resource %s: value must be a hash", resname);

    const char* svctype = NULL;
    const vscf_data_t* svctype_data = vscf_hash_get_data_byconstkey(opts, "service_type", true);
    if(svctype_data) {
        if(VSCF_SIMPLE_T != vscf_get_type(svctype_data))
            log_fatal("plugin_simplefo: resource %s: 'service_type' must be a string", resname);
        svctype = vscf_simple_get_data(svctype_data);
    }

    const vscf_data_t* primary_data = vscf_hash_get_data_byconstkey(opts, "primary", true);
    if(!primary_data || VSCF_SIMPLE_T != vscf_get_type(primary_data))
        log_fatal("plugin_simplefo: resource %s: 'primary' must be defined as an IPv4 address string", resname);
    {
        struct in_addr a;
        const char* addr_txt = vscf_simple_get_data(primary_data);
        if(inet_pton(AF_INET, addr_txt, &a) < 1)
            log_fatal("plugin_simplefo: resource %s: parsing '%s' as IPv4 address failed", resname, addr_txt);
        this_res->primary_ipv4 = a.s_addr;
        char* desc = malloc(strlen(resname) + 5);
        strcpy(desc, resname);
        strcat(desc, "/pri");
        monio_add(svctype, desc, addr_txt, &this_res->primary_state);
    }

    const vscf_data_t* secondary_data = vscf_hash_get_data_byconstkey(opts, "secondary", true);
    if(!secondary_data || VSCF_SIMPLE_T != vscf_get_type(secondary_data))
        log_fatal("plugin_simplefo: resource %s: 'secondary' must be defined as an IPv4 address string", resname);
    {
        struct in_addr a;
        const char* addr_txt = vscf_simple_get_data(secondary_data);
        if(inet_pton(AF_INET, addr_txt, &a) < 1)
            log_fatal("plugin_simplefo: resource %s: parsing '%s' as IPv4 address failed", resname, addr_txt);
        this_res->secondary_ipv4 = a.s_addr;
        char* desc = malloc(strlen(resname) + 5);
        strcpy(desc, resname);
        strcat(desc, "/sec");
        monio_add(svctype, desc, addr_txt, &this_res->secondary_state);
    }

    vscf_hash_iterate(opts, true, bad_res_opt, (void*)resname);
    return true;
}

static void log_stats(void) {
    unsigned long pri = 0;
    unsigned long sec = 0;
    for(unsigned i = 0; i < n_threads; i++) {
        pri += satom_get(&thread_stats[i]->pri);
        sec += satom_get(&thread_stats[i]->sec);
    }
    log_info("plugin_simplefo: runtime queries answered with primary addrs: %lu secondary addrs: %lu", pri, sec);
}

F_NONNULL
static void log_watcher_cb(struct ev_loop* loop V_UNUSED, ev_timer* t V_UNUSED, int revents V_UNUSED) {
    dmn_assert(loop); dmn_assert(t);
    log_stats();
}

/*********************************/
/* Exported callbacks start here */
/*********************************/

monio_list_t* plugin_simplefo_load_config(const vscf_data_t* config) {
    dmn_assert(config);
    dmn_assert(vscf_get_type(config) == VSCF_HASH_T);

    const vscf_data_t* logival_data = vscf_hash_get_data_byconstkey(config, "log_interval", true);
    if(logival_data &&
       (VSCF_SIMPLE_T != vscf_get_type(logival_data)
       || !vscf_simple_get_as_ulong(logival_data, &log_interval))
    )
        log_fatal("plugin_simplefo: 'log_interval' must be an unsigned integer if defined");

    const vscf_data_t* res_data = vscf_hash_get_data_byconstkey(config, "resources", true);
    if(!res_data || VSCF_HASH_T != vscf_get_type(res_data))
        log_fatal("plugin_simplefo: 'resources' must exist and must be a hash");

    num_resources = vscf_hash_get_len(res_data);
    resources = malloc(num_resources * sizeof(sfo_resource_t));
    unsigned residx = 0;
    vscf_hash_iterate(res_data, false, config_res, &residx);

    return &monio_list;
}

void plugin_simplefo_full_config(unsigned num_threads) {
    n_threads = num_threads;
    thread_stats = calloc(num_threads, sizeof(sfo_stats_t*));
}

unsigned plugin_simplefo_map_resource_dyna(const char* resname) {
    if(!resname)
        log_fatal("simplefo plugin requires a resource name");

    for(unsigned i = 0; i < num_resources; i++)
        if(!strcmp(resname, resources[i].name))
            return i;

    log_fatal("Unknown simplefo plugin resource '%s'", resname);
}

void plugin_simplefo_pre_privdrop(void) {
    runtime = true;
}

void plugin_simplefo_pre_run(struct ev_loop* loop) {
    // Set up the log_stats() callback to fire every half hour
    //  in the main thread.
    ev_timer* log_watcher = malloc(sizeof(ev_timer));
    ev_timer_init(log_watcher, log_watcher_cb, log_interval, log_interval);
    ev_set_priority(log_watcher, -2);
    ev_timer_start(loop, log_watcher);
}

void plugin_simplefo_iothread_init(unsigned threadnum) {
    thread_stats[threadnum] = calloc(1, sizeof(sfo_stats_t));
}

void plugin_simplefo_resolve_dynaddr(unsigned threadnum, unsigned resnum, const anysin_t* client V_UNUSED, dynaddr_result_t* result) {
    result->count_v6 = 0;
    result->count_v4 = 1;

    // Non-runtime (early, internal packet-sizing queries) don't
    //  care about correct address data, just correct counts, and
    //  this plugin always returns exactly 1 IPv4 and 0 IPv6.
    if(!runtime) return;
    
    sfo_resource_t* res = &resources[resnum];
    monio_state_uint_t p_state = monio_state_get(&res->primary_state);

    bool use_primary = true;

    switch(p_state) {
        case MONIO_STATE_UP:
            break;
        case MONIO_STATE_DANGER:
            result->ttl >>= 1; // Cut TTL in half
            break;
        case MONIO_STATE_DOWN:
            result->ttl >>= 1; // Cut TTL in half
            monio_state_uint_t s_state = monio_state_get(&res->secondary_state);
            if(s_state != MONIO_STATE_DOWN)
                use_primary = false;
            break;
        default:
            dmn_assert(0);
    }

    if(use_primary) {
        result->addrs_v4[0] = res->primary_ipv4;
        satom_inc(&thread_stats[threadnum]->pri);
    }
    else {
        result->addrs_v4[0] = res->secondary_ipv4;
        satom_inc(&thread_stats[threadnum]->sec);
    }
}

void plugin_simplefo_exit(void) {
    log_info("plugin_simplefo: final stats:");
    log_stats();
}
